@@ -81,7 +81,7 @@ static std::vector<string> get_args(const PERL_CONTEXT* cx) {
 
 static PerlTraceSP get_trace() noexcept {
     dTHX;
-    std::vector<StackframePtr> frames;
+    std::vector<StackframeSP> frames;
     I32 level = 0;
     const PERL_CONTEXT *dbcx = nullptr;
     const PERL_CONTEXT* cx = caller_cx(level, &dbcx);
@@ -112,7 +112,7 @@ static PerlTraceSP get_trace() noexcept {
 
         auto args = get_args(cx);
 
-        StackframePtr frame(new Stackframe());
+        StackframeSP frame(new Stackframe());
         frame->library = library;
         frame->file = file;
         frame->line_no = line;
